"""Models module with the training loop and logic to handle data which feeds into the loop."""

from ._models import Model
from .nuclei import SegmentationModel
from .puncta import SpotsModel

__all__ = ["Model", "segmentation", "puncta_detection"]